import math
import ctypes.util
import warnings
from ctypes import *

import numba
from numba.minivect import minitypes
from numba.minivect.minitypes import *
import numba.utils

#-------------------------------------------------------------------
# CTypes Types for Type Checking
#-------------------------------------------------------------------

# ______________________________________________________________________
# Meta types

_ctypes_func_type = type(ctypes.CFUNCTYPE(ctypes.c_int))
_ctypes_pointer_type = type(ctypes.POINTER(ctypes.c_int))
_ctypes_array_type = type(ctypes.c_int * 2)

ctypes_map = {
    c_char : char,
    c_byte : char,
    c_ubyte : uchar,
    c_short : short,
    c_ushort : ushort,
    c_int : int_,
    c_uint : uint,
    c_long : c_ulong,
    c_long : long_,
    c_ulong : ulong,
    c_longlong : longlong,
    c_ulonglong : ulonglong,

    c_float : float_,
    c_double : double,

    c_char_p : c_string_type,
    c_void_p : void.pointer(),
    None : void,

    py_object : object_,
    }

# On MSVC Windows, these are equal
if c_longdouble != c_double:
    ctypes_map[c_longdouble] = longdouble


# ______________________________________________________________________
# Value types

CData = type(ctypes.c_int(10)).__mro__[-2]

#-------------------------------------------------------------------
# Check Whether values are ctypes values
#-------------------------------------------------------------------

def is_ctypes_function(value):
    return isinstance(type(value), _ctypes_func_type)

def is_ctypes_value(ctypes_value):
    return isinstance(ctypes_value, CData)

def is_ctypes_struct_type(ctypes_type):
    return (isinstance(ctypes_type, type) and
            issubclass(ctypes_type, ctypes.Structure))

def is_ctypes_type(ctypes_type):
    return (
       (numba.utils.hashable(ctypes_type) and ctypes_type in ctypes_map) or
       is_ctypes_struct_type(ctypes_type)
    )

def is_ctypes(value):
    "Check whether the given value is a ctypes value"
    return is_ctypes_value(value) or is_ctypes_type(value)

#-------------------------------------------------------------------
# Type mapping
#-------------------------------------------------------------------

def from_ctypes_type(ctypes_type):
    """
    Convert a ctypes type to a numba type
    """
    if numba.utils.hashable(ctypes_type) and ctypes_type in ctypes_map:
        return ctypes_map[ctypes_type]
    elif isinstance(ctypes_type, _ctypes_pointer_type):
        return from_ctypes_type(ctypes_type._type_).pointer()
    elif isinstance(ctypes_type, _ctypes_array_type):
        base_type = from_ctypes_type(ctypes_type._type_)
        return minitypes.CArrayType(base_type, ctypes_type._length_)
    elif issubclass(ctypes_type, ctypes.Structure):
        fields = [(name, from_ctypes_type(field_type))
                      for name, field_type in ctypes_type._fields_]
        return numba.struct(fields)
    else:
        raise NotImplementedError(ctypes_type)

def from_ctypes_value(value):
    """
    Convert a ctypes value to a numba type
    """
    from numba import typesystem

    if is_ctypes_type(value):
        # Value is a ctypes type, e.g. c_int
        return typesystem.CastType(from_ctypes_type(value))

    elif is_ctypes_function(value):
        # TODO: move this to from_ctypes_type
        if value.argtypes is None:
            warnings.warn(
                "ctypes function %s has no argument types set" % (value,))
            return object_

        restype = from_ctypes_type(value.restype)
        argtypes = [from_ctypes_type(at) for at in value.argtypes]
        signature = minitypes.FunctionType(return_type=restype, args=argtypes)
        return signature

    elif is_ctypes_type(type(value)) or hasattr(value, '_type_'):
        # Value is a ctypes value, e.g. c_int(10)
        result_type = from_ctypes_type(type(value))

        if result_type.is_pointer:
            # Handle ctypes pointers
            try:
                ctypes.cast(value, ctypes.c_void_p)
            except ctypes.ArgumentError:
                pass
            else:
                addr_int = ctypes.cast(value, ctypes.c_void_p).value
                result_type = typesystem.KnownPointerType(
                    result_type.base_type, addr_int)

        return result_type

    else:
        raise NotImplementedError(value)



import numba
from numba import *

@autojit
def test_object_iteration(obj):
    """
    >>> test_object_iteration([1, 2, 3])
    1
    2
    3
    """
    for x in obj:
        print(x)

if __name__ == '__main__':
#    test_object_iteration([1, 2, 3])
    numba.testing.testmod()



# -*- coding: utf-8 -*-
"""
Generate generated_conversions.c

Utilities adjusted from Cython/Compiler/PyrexTypes.pyx
"""
from __future__ import print_function, division, absolute_import

import os

func_name = "__Numba_PyInt_As%(SignWord)s%(TypeName)s"
header = "static %%(type)s %(FuncName)s(PyObject* x)" % { 'FuncName' : func_name}

c_int_from_py_function = """
%(Header)s {
    const %(type)s neg_one = (%(type)s)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
    if (sizeof(%(type)s) < sizeof(long)) {
        long val = __Numba_PyInt_AsSignedLong(x);
        if (unlikely(val != (long)(%(type)s)val)) {
            if (!unlikely(val == -1 && PyErr_Occurred())) {
                PyErr_SetString(PyExc_OverflowError,
                    (is_unsigned && unlikely(val < 0)) ?
                    "can't convert negative value to %(type)s" :
                    "value too large to convert to %(type)s");
            }
            return (%(type)s)-1;
        }
        return (%(type)s)val;
    }
    return (%(type)s)__Numba_PyInt_As%(SignWord)sLong(x);
}
"""

c_long_from_py_function = """
%(Header)s {
    const %(type)s neg_one = (%(type)s)-1, const_zero = 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_VERSION_HEX < 0x03000000
    if (likely(PyInt_Check(x))) {
        long val = PyInt_AS_LONG(x);
        if (is_unsigned && unlikely(val < 0)) {
            PyErr_SetString(PyExc_OverflowError,
                            "can't convert negative value to %(type)s");
            return (%(type)s)-1;
        }
        return (%(type)s)val;
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
            if (unlikely(Py_SIZE(x) < 0)) {
                PyErr_SetString(PyExc_OverflowError,
                                "can't convert negative value to %(type)s");
                return (%(type)s)-1;
            }
            return (%(type)s)PyLong_AsUnsigned%(TypeName)s(x);
        } else {
            return (%(type)s)PyLong_As%(TypeName)s(x);
        }
    } else {
        %(type)s val;
        PyObject *tmp = __Numba_PyNumber_Int(x);
        if (!tmp) return (%(type)s)-1;
        val = __Numba_PyInt_As%(SignWord)s%(TypeName)s(tmp);
        Py_DECREF(tmp);
        return val;
    }
}
"""

def rank(types):
    types = [type for name, type in types]
    return dict(zip(types, range(len(exact_types))))

# Builtin C types that we know how to convert to/from objects
exact_types = (
    ("Char",     "char"),
    ("Short",    "short"),
    ("Int",      "int"),
    ("Long",     "long"),
    ("LongLong", "PY_LONG_LONG"),
)
rank_exact = rank(exact_types)

# Types for which we don't know the mapping to exact_types
inexact_types = (
    ("Py_ssize_t", "Py_ssize_t"),
    ("size_t",     "size_t"),
    ("npy_intp",   "npy_intp"),
)
rank_inexact = rank(inexact_types)

signednesses = (
    "signed",
    "unsigned",
)

def write_utility(exact_type, exact_type_name, out_c, out_h, signedness):
    # Select utility template
    if rank_exact[exact_type] < rank_exact["long"]:
        utility = c_int_from_py_function
    else:
        utility = c_long_from_py_function

    # Build argument dict
    fmtargs = { 'TypeName' : exact_type_name,
                'SignWord' : signedness.title(),
                'type'     : signedness + " " + exact_type }
    fmtargs.update(FuncName=func_name % fmtargs,
                   Header=header % fmtargs)

    # Write results
    conversion = utility % fmtargs
    out_c.write(conversion)
    out_h.write(header % fmtargs + ';\n')

    # print_export(fmtargs)
    # print_utility_load(fmtargs, signedness)


def generate_conversions(out_c, out_h):
    "Generate numba/external/utilities/generated_conversions.c"
    out_c.write("/* This file is generated by %s, do not edit */\n" %
                                           os.path.abspath(__file__))
    out_c.write('#include "%s"\n\n' % out_h.name)

    for exact_type_name, exact_type in exact_types:
        for signedness in signednesses:
            write_utility(exact_type, exact_type_name, out_c, out_h, signedness)

    # write_utility("char", "Char", out_c, out_h, "signed")

    print("Wrote %s and %s" % (out_c.name, out_h.name))


def print_export(fmtargs):
    "Code to put in type_conversion.c"
    print("EXPORT_FUNCTION(%(FuncName)s, module, error)" % fmtargs)

def print_utility_load(fmtargs, signedness):
    "Code to put in numba.external.utility"
    typename = fmtargs['TypeName'].lower()
    if signedness == "unsigned":
        typename = "u" + typename

    print('%-10s : load("%s", %s(object_)),' % (typename, fmtargs['FuncName'],
                                                typename))


def open_files():
    numba_root = os.path.dirname(os.path.abspath(__file__))
    root = os.path.join(numba_root, "numba", "external", "utilities")

    out_c = open(os.path.join(root, "generated_conversions.c"), "w")
    out_h = open(os.path.join(root, "generated_conversions.h"), "w")
    return out_c, out_h

def run():
    generate_conversions(*open_files())

if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

from numba.minivect.minitypes import *
from numba.minivect.minitypes import (FunctionType)

from .basetypes import *
from .closuretypes import *
from .ssatypes import *
from .templatetypes import *
from .containertypes import *
from .typemapper import *
from .typeutils import *

from .shorthands import *

# from typeset import *
from .typematch import *

__all__ = minitypes.__all__ + [
    'O', 'b1', 'i1', 'i2', 'i4', 'i8', 'u1', 'u2', 'u4', 'u8',
    'f4', 'f8', 'f16', 'c8', 'c16', 'c32', 'template',
]



import numpy as np

import numba
from numba import *
from numba import typesystem

tup_t = typesystem.TupleType

#------------------------------------------------------------------------
# Test functions
#------------------------------------------------------------------------

@autojit
def array(value):
    return numba.typeof(np.array(value))

@autojit
def nonzero(value):
    return numba.typeof(np.nonzero(value))

@autojit
def where(value):
    return numba.typeof(np.where(value))

@autojit
def where3(value, x, y):
    return numba.typeof(np.where(value, x, y))

@autojit
def numba_dot(A, B):
    result = np.dot(A, B)
    return numba.typeof(result), result

@autojit
def numba_vdot(A, B):
    result = np.vdot(A, B)
    return numba.typeof(result), result

@autojit
def numba_inner(a, b):
    result = np.inner(a, b)
    return numba.typeof(result), result

@autojit
def numba_outer(a, b):
    result = np.outer(a, b)
    return numba.typeof(result), result

@autojit
def numba_tensordot(a, b, axes):
    result = np.tensordot(a, b, axes)
    return numba.typeof(result), result

@autojit
def numba_tensordot2(a, b):
    result = np.tensordot(a, b)
    return numba.typeof(result), result

@autojit
def numba_kron(a, b):
    result = np.kron(a, b)
    return numba.typeof(result), result

# ------------- Test sum ------------

@autojit
def sum_(a):
    return numba.typeof(np.sum(a))

@autojit
def sum_axis(a, axis):
    return numba.typeof(np.sum(a, axis=axis))

@autojit
def sum_dtype(a, dtype):
    return numba.typeof(np.sum(a, dtype=dtype))

@autojit
def sum_out(a, out):
    return numba.typeof(np.sum(a, out=out))

#------------------------------------------------------------------------
# Tests
#------------------------------------------------------------------------

def equals(a, b):
    assert a == b, (a, b, type(a), type(b),
                    a.comparison_type_list, b.comparison_type_list)

def test_array():
    equals(array(np.array([1, 2, 3], dtype=np.double)), float64[:])
    equals(array(np.array([[1, 2, 3]], dtype=np.int32)), int32[:, :])
    equals(array(np.array([[1, 2, 3],
                           [4, 5, 6]], dtype=np.int32).T), int32[:, :])

def test_nonzero():
    equals(nonzero(np.array([1, 2, 3], dtype=np.double)),
           tup_t(npy_intp[:], 1))
    equals(nonzero(np.array([[1, 2, 3]], dtype=np.double)),
           tup_t(npy_intp[:], 2))
    equals(nonzero(np.array((((1, 2, 3),),), dtype=np.double)),
           tup_t(npy_intp[:], 3))

def test_where():
    equals(where(np.array([1, 2, 3], dtype=np.double)),
           tup_t(npy_intp[:], 1))

    equals(where3(np.array([True, False, True]),
                  np.array([1, 2, 3], dtype=np.double),
                  np.array([1, 2, 3], dtype=np.complex128)),
           complex128[:])

    equals(where3(np.array([True, False, True]),
                  np.array([1, 2, 3], dtype=np.float32),
                  np.array([1, 2, 3], dtype=np.int64)),
           float64[:])

def test_numba_dot():
    A = np.array(1)
    B = np.array(2)

    dtype = typesystem.from_numpy_dtype(A.dtype).dtype

    for i in range(1, 5):
        for j in range(1, 5):
            # print i, j

            shape_A = (1,) * i
            shape_B = (1,) * j

            x = A.reshape(*shape_A)
            y = B.reshape(*shape_B)

            result_type, result = numba_dot(x, y)

            assert result == np.dot(x, y)
            if i + j - 2 > 0:
                assert result.ndim == result_type.ndim
            else:
                assert result_type == dtype

def test_numba_vdot():
    for a, b in ((np.array([1+2j,3+4j]),
                  np.array([5+6j,7+8j])),
                 (np.array([[1, 4], [5, 6]]),
                  np.array([[4, 1], [2, 2]]))):
        result_type, result = numba_vdot(a, b)
        assert result == np.vdot(a, b)
        assert result_type == typesystem.from_numpy_dtype(a.dtype).dtype
        result_type, result = numba_vdot(b, a)
        assert result == np.vdot(b, a)
        assert result_type == typesystem.from_numpy_dtype(b.dtype).dtype

def test_numba_inner():
    # Note these tests assume that the lhs' type is the same as the
    # promotion type for both arguments.  They will fail if additional
    # test data doesn't adhere to this policy.
    for a, b in ((np.array([1,2,3]), np.array([0,1,0])),
                 (np.arange(24).reshape((2,3,4)), np.arange(4)),
                 (np.eye(2), 7)):
        result_type, result = numba_inner(a, b)
        if result_type.is_array:
            assert (result == np.inner(a, b)).all()
            assert (result_type.dtype ==
                    typesystem.from_numpy_dtype(result.dtype).dtype)
            assert (result_type.dtype ==
                    typesystem.from_numpy_dtype(a.dtype).dtype)
        else:
            assert result == np.inner(a, b)
            assert result_type == typesystem.from_numpy_dtype(a.dtype).dtype

def test_numba_outer():
    for a, b in ((np.ones((5,)), np.linspace(-2, 2, 5)),
                 (1j * np.linspace(2, -2, 5), np.ones((5,))),
                 (np.array(['a', 'b', 'c'], dtype=object), np.arange(1,4)),
                 (np.array([1]), 1),
                 (np.ones((2,2,2)), np.linspace(-2, 2, 5))):
        result_type, result = numba_outer(a, b)
        assert (result == np.outer(a, b)).all()
        assert (result_type.is_array and result_type.ndim == 2)
        assert result_type.dtype == typesystem.from_numpy_dtype(a.dtype).dtype

def test_numba_tensordot():
    for a, b, axes in ((np.arange(60.).reshape(3, 4, 5),
                        np.arange(24.).reshape(4, 3, 2), ([1,0],[0,1])),
                       ):
        result_type, result = numba_tensordot(a, b, axes)
        assert (result == np.tensordot(a, b, axes)).all()
        # See comments in the docstring for
        # numba.type_inference.modules.numpymodule.tensordot().
        assert result_type == object_

def test_numba_tensordot2():
    A = np.array(1)
    B = np.array(2)

    dtype = typesystem.from_numpy_dtype(A.dtype).dtype
    for i in range(2, 5):
        for j in range(2, 5):

            shape_A = (1,) * i
            shape_B = (1,) * j

            x = A.reshape(*shape_A)
            y = B.reshape(*shape_B)

            result_type, result = numba_tensordot2(x, y)
            control = np.tensordot(x, y)
            assert result == control
            #assert result_type == numba.typeof(control)
            if i + j - 4 > 0:
                assert result.ndim == result_type.ndim
            else:
                assert result_type == dtype

def test_sum():
    a = np.array([1, 2, 3], dtype=np.int32)
    b = np.array([[1, 2], [3, 4]], dtype=np.int64)

    equals(sum_(a), int32)
    equals(sum_axis(a, 0), int32)
    equals(sum_dtype(a, np.double), double)
    equals(sum_out(b, a), int32[:]) # Not a valid call to sum :)

if __name__ == "__main__":
    test_array()
    test_nonzero()
    test_where()
    test_numba_dot()
    test_numba_vdot()
    test_numba_inner()
    test_numba_outer()
    test_numba_tensordot()
    test_numba_tensordot2()
    test_sum()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
from numba import *
from .intrinsic import Intrinsic
from llpython.byte_translator import LLVMTranslator

__all__ = ['CStringSlice2',
           'CStringSlice2Len']

class CStringSlice2 (Intrinsic):
    arg_types = [c_string_type, c_string_type, size_t, Py_ssize_t, Py_ssize_t]
    return_type = void

    def implementation(self, module, lfunc):
        # logger.debug((module, str(lfunc)))
        def _py_c_string_slice(out_string, in_string, in_str_len, lower,
                               upper):
            zero = lc_size_t(0)
            if lower < zero:
                lower += in_str_len
            if upper < zero:
                upper += in_str_len
            elif upper > in_str_len:
                upper = in_str_len
            temp_len = upper - lower
            if temp_len < zero:
                temp_len = zero
            strncpy(out_string, in_string + lower, temp_len)
            out_string[temp_len] = li8(0)
            return
        LLVMTranslator(module).translate(_py_c_string_slice,
                                         llvm_function = lfunc)
        return lfunc

class CStringSlice2Len(Intrinsic):
    arg_types = [c_string_type, size_t, Py_ssize_t, Py_ssize_t]
    return_type = size_t

    def implementation(self, module, lfunc):
        def _py_c_string_slice_len(in_string, in_str_len, lower, upper):
            zero = lc_size_t(0)
            if lower < zero:
                lower += in_str_len
            if upper < zero:
                upper += in_str_len
            elif upper > in_str_len:
                upper = in_str_len
            temp_len = upper - lower
            if temp_len < zero:
                temp_len = zero
            return temp_len + lc_size_t(1)
        LLVMTranslator(module).translate(_py_c_string_slice_len,
                                         llvm_function = lfunc)
        return lfunc



#! /usr/bin/env python
# ______________________________________________________________________
'''test_fbcorr

Test the fbcorr() example .... 
'''
# ______________________________________________________________________

import numpy as np
import numba
from numba.decorators import jit
nd4type = numba.double[:,:,:,:]

import sys
import unittest

# ______________________________________________________________________

def fbcorr(imgs, filters, output):
    n_imgs, n_rows, n_cols, n_channels = imgs.shape
    n_filters, height, width, n_ch2 = filters.shape

    for ii in range(n_imgs):
        for rr in range(n_rows - height + 1):
            for cc in range(n_cols - width + 1):
                for hh in xrange(height):
                    for ww in xrange(width):
                        for jj in range(n_channels):
                            for ff in range(n_filters):
                                imgval = imgs[ii, rr + hh, cc + ww, jj]
                                filterval = filters[ff, hh, ww, jj]
                                output[ii, ff, rr, cc] += imgval * filterval

# ______________________________________________________________________

class TestFbcorr(unittest.TestCase):
    def test_vectorized_fbcorr(self):
        ufbcorr = jit(argtypes=(nd4type, nd4type, nd4type))(fbcorr)

        imgs = np.random.randn(10, 16, 16, 3)
        filt = np.random.randn(6, 5, 5, 3)
        old_output = np.zeros((10, 6, 15, 15))
        fbcorr(imgs, filt, old_output)
        new_output = np.zeros((10, 6, 15, 15))
        ufbcorr(imgs, filt, new_output)

        self.assertTrue((abs(old_output - new_output) < 1e-9).all())

    def test_catch_error(self):
        imgs = np.random.randn(10, 64, 64, 3)
        filt = np.random.randn(6, 5, 5, 3)
        #incorrect channel-minor format?
        old_output = np.zeros((10, 60, 60, 6))

        try:
            fbcorr(imgs, filt, old_output)
        except IndexError as e:
            print(('This test produced the error "' + repr(e) + '"'))
        else:
            raise Exception('This should have produced an error.')

# ______________________________________________________________________

if __name__ == "__main__":
    unittest.main()

# ______________________________________________________________________
# End of test_filter2d.py



"""
>>> test_typeof()
"""

import numba
from numba import *

def make_base(compiler):
    @compiler
    class Base(object):

        value1 = double
        value2 = int_

        @void(int_, double)
        def __init__(self, value1, value2):
            self.value1 = value1
            self.value2 = value2

    return Base

Base = make_base(jit)

@jit
class Derived(Base):

    value3 = float_

    @void(int_)
    def setvalue(self, value):
        self.value3 = value

@autojit
def base_typeof():
    obj1 = Base(10, 11.0)
    return numba.typeof(obj1.value1), numba.typeof(obj1.value2)

@autojit
def derived_typeof():
    obj = Derived(10, 11.0)
    return (numba.typeof(obj.value1),
            numba.typeof(obj.value2),
            numba.typeof(obj.value3))

def test_typeof():
    pass
    # TODO: type recognition of extension object instantiation
    # assert base_typeof() == (double, int_), base_typeof()
    # assert derived_typeof() == (double, int_, float_), derived_typeof()


#------------------------------------------------------------------------
# Test Specialized autojit typeof
#------------------------------------------------------------------------

AutoBase = make_base(autojit)

@autojit
def attrtypes(obj):
    return numba.typeof(obj.value1), numba.typeof(obj.value2)

def test_autobase():
    obj = AutoBase(10, 11.0)
    assert attrtypes(obj) == (double, int_)

if __name__ == '__main__':
    test_typeof()
    test_autobase()



import os
import ctypes

from numba import *
import numba

try:
    import cffi
    ffi = cffi.FFI()
except ImportError:
    ffi = None

root = os.path.dirname(os.path.abspath(__file__))
include_dirs = [root]

# ______________________________________________________________________

def test():
    if ffi is not None:
        run_callback()

# ______________________________________________________________________
# Tests

@jit(int_(int_, int_))
def numba_callback(a, b):
    return a * b

@autojit #(nopython=True)
def call_cffi_func(eat_callback):
    return eat_callback(numba_callback)

def run_callback():
    # Define C function taking a callback
    ffi.cdef("typedef int (*callback_t)(int, int);")
    ffi.cdef("typedef int (*eat_callback_t)(callback_t);")
    ffi.cdef("eat_callback_t get_eat_callback();", override=True)

    ffi.cdef("int printf(char *, ...);")
    lib = ffi.verify("""
        typedef int (*callback_t)(int, int);
        typedef int (*eat_callback_t)(callback_t);

        int eat_callback(callback_t callback) {
            return callback(5, 2);
        }

        eat_callback_t get_eat_callback() {
            return (callback_t) eat_callback;
        }
    """, include_dirs=include_dirs)

    # CFFI returns builtin methods instead of CData functions for
    # non-external functions. Get the CDATA function through an indirection.
    eat_callback = lib.get_eat_callback()

    assert call_cffi_func(eat_callback) == 10


# ______________________________________________________________________

if __name__ == "__main__":
    test()
